"""
This module provides a Python implementation of the `main_loop_api` of libpulse to use a Python asyncio event loop for
libpulse's asynchronous event handling. This libpulse main loop implementation can be used as an alternative to
libpulse's default mainloop, when setting up a libpulse *context* object via Python ctypes bindings.

Usage::

    loop = asyncio.get_event_loop()
    main_loop = PythonMainLoop(loop)

    context = libpulse.context_new(main_loop.api_pointer, b"my client name")

Copyright (c) 2021 Michael Thies
"""

import asyncio
import ctypes as c
import itertools
import sysconfig
import time
from functools import partial
from typing import Set, Optional, Dict
from pulsectl._pulsectl import PA_MAINLOOP_API

# References:
# - https://docs.python.org/3.9/library/ctypes.html
# - https://freedesktop.org/software/pulseaudio/doxygen/mainloop-api_8h.html


class pa_mainloop_api(PA_MAINLOOP_API):
    pass


# Find out the platform specific type of time_t
time_t_size = sysconfig.get_config_var('SIZEOF_TIME_T')
if time_t_size == sysconfig.get_config_var('SIZEOF_LONG_LONG'):
    time_t = c.c_longlong
elif time_t_size == sysconfig.get_config_var('SIZEOF_LONG'):
    time_t = c.c_long
else:
    raise RuntimeError("Current platform has unexpected sizeof(time_t)")


class timeval(c.Structure):
    _fields_ = [
        ('tv_sec', time_t),
        ('tv_usec', c.c_long),
    ]

    def to_float(self) -> float:
        return self.tv_sec + self.tv_usec / 1e6


PA_IO_EVENT_NULL = 0
PA_IO_EVENT_INPUT = 1
PA_IO_EVENT_OUTPUT = 2
PA_IO_EVENT_HANGUP = 4
PA_IO_EVENT_ERROR = 8

# Typedefs from mainloop.h
pa_defer_event_p = c.c_void_p
pa_defer_event_cb_t = c.CFUNCTYPE(None, c.POINTER(pa_mainloop_api), pa_defer_event_p, c.c_void_p)
pa_defer_event_destroy_cb_t = c.CFUNCTYPE(None, c.POINTER(pa_mainloop_api), pa_defer_event_p, c.c_void_p)
pa_io_event_p = c.c_void_p
pa_io_event_flags = c.c_int
pa_io_event_cb_t = c.CFUNCTYPE(None, c.POINTER(pa_mainloop_api), pa_io_event_p, c.c_int, pa_io_event_flags, c.c_void_p)
pa_io_event_destroy_cb_t = c.CFUNCTYPE(None, c.POINTER(pa_mainloop_api), pa_io_event_p, c.c_void_p)
pa_time_event_p = c.c_void_p
pa_time_event_cb_t = c.CFUNCTYPE(None, c.POINTER(pa_mainloop_api), pa_time_event_p, c.POINTER(timeval), c.c_void_p)
pa_time_event_destroy_cb_t = c.CFUNCTYPE(None, c.POINTER(pa_mainloop_api), pa_time_event_p, c.c_void_p)

# function pointer types of pa_mainloop_api struct
pa_io_new_t = c.CFUNCTYPE(pa_io_event_p, c.POINTER(pa_mainloop_api), c.c_int, pa_io_event_flags, pa_io_event_cb_t,
                          c.c_void_p)
pa_io_enable_t = c.CFUNCTYPE(None, pa_io_event_p, pa_io_event_flags)
pa_io_set_destroy_t = c.CFUNCTYPE(None, pa_io_event_p, pa_io_event_destroy_cb_t)
pa_io_free_t = c.CFUNCTYPE(None, pa_io_event_p)
pa_time_new_t = c.CFUNCTYPE(pa_time_event_p, c.POINTER(pa_mainloop_api), c.POINTER(timeval), pa_time_event_cb_t,
                            c.c_void_p)
pa_time_restart_t = c.CFUNCTYPE(None, pa_time_event_p, c.POINTER(timeval))
pa_time_set_destroy_t = c.CFUNCTYPE(None, pa_time_event_p, pa_time_event_destroy_cb_t)
pa_time_free_t = c.CFUNCTYPE(None, pa_time_event_p)
pa_defer_new_t = c.CFUNCTYPE(pa_defer_event_p, c.POINTER(pa_mainloop_api), pa_defer_event_cb_t, c.c_void_p)
pa_defer_enable_t = c.CFUNCTYPE(None, pa_defer_event_p, c.c_bool)
pa_defer_free_t = c.CFUNCTYPE(None, pa_defer_event_p)
pa_defer_set_destroy_t = c.CFUNCTYPE(None, pa_defer_event_p, pa_defer_event_destroy_cb_t)
pa_quit_t = c.CFUNCTYPE(None, c.POINTER(pa_mainloop_api), c.c_int)

pa_mainloop_api._fields_ = [
    ("userdata", c.c_void_p),  # We use it to store a pointer to the corresponding PythonMainLoop python object
    ("io_new", pa_io_new_t),
    ("io_enable", pa_io_enable_t),
    ("io_free", pa_io_free_t),
    ("io_set_destroy", pa_io_set_destroy_t),
    ("time_new", pa_time_new_t),
    ("time_restart", pa_time_restart_t),
    ("time_free", pa_time_free_t),
    ("time_set_destroy", pa_time_set_destroy_t),
    ("defer_new", pa_defer_new_t),
    ("defer_enable", pa_defer_enable_t),
    ("defer_free", pa_defer_free_t),
    ("defer_set_destroy", pa_defer_set_destroy_t),
    ("quit", pa_quit_t),
]


class PythonMainLoop:
    __slots__ = ('loop', 'io_events', 'time_events', 'defer_events',
                 'api_pointer', 'io_reader_events', 'io_writer_events', 'retval')

    def __init__(self, loop: asyncio.AbstractEventLoop):
        # TODO implement 'on_quit' event callback
        self.loop = loop
        self.io_events: Set["PythonIOEvent"] = set()
        self.defer_events: Set["PythonDeferEvent"] = set()
        self.time_events: Set["PythonTimeEvent"] = set()
        self.io_reader_events: Dict[int, Set[PythonIOEvent]] = {}
        self.io_writer_events: Dict[int, Set[PythonIOEvent]] = {}
        self.api_pointer = c.pointer(self._create_api())
        self.retval: Optional[int] = None

    def _create_api(self) -> pa_mainloop_api:
        result = pa_mainloop_api()
        result.userdata = c.cast(c.pointer(c.py_object(self)), c.c_void_p)
        result.io_new = aio_io_new
        result.io_enable = aio_io_enable
        result.io_free = aio_io_free
        result.io_set_destroy = aio_io_set_destroy
        result.time_new = aio_time_new
        result.time_restart = aio_time_restart
        result.time_free = aio_time_free
        result.time_set_destroy = aio_time_set_destroy
        result.defer_new = aio_defer_new
        result.defer_enable = aio_defer_enable
        result.defer_free = aio_defer_free
        result.defer_set_destroy = aio_defer_set_destroy
        result.quit = aio_quit
        return result

    def register_unregister_io_event(self, event: "PythonIOEvent", reader: bool, writer: bool) -> None:
        if writer:
            if event.fd in self.io_writer_events:
                self.io_writer_events[event.fd].add(event)
            else:
                self.io_writer_events[event.fd] = {event}
                self.loop.add_writer(event.fd, partial(self._io_write_callback, event.fd))
        elif event.fd in self.io_writer_events:
            self.io_writer_events[event.fd].discard(event)
            if not self.io_writer_events[event.fd]:
                del self.io_writer_events[event.fd]
                self.loop.remove_writer(event.fd)

        if reader:
            if event.fd in self.io_reader_events:
                self.io_reader_events[event.fd].add(event)
            else:
                self.io_reader_events[event.fd] = {event}
                self.loop.add_reader(event.fd, partial(self._io_read_callback, event.fd))
        elif event.fd in self.io_reader_events:
            self.io_reader_events[event.fd].discard(event)
            if not self.io_reader_events[event.fd]:
                del self.io_reader_events[event.fd]
                self.loop.remove_reader(event.fd)
        # Python asyncio's API does not allow us to `poll` for HANGUP and ERROR states.
        # However, this does not seem to be an issue, since the pulse client library obviously does not use these flags
        # for io io_events.

    def _io_write_callback(self, fd) -> None:
        for event in tuple(self.io_writer_events.get(fd, ())):
            event.write()

    def _io_read_callback(self, fd) -> None:
        for event in tuple(self.io_reader_events.get(fd, ())):
            event.read()

    def stop(self, retval: int) -> None:
        for event in itertools.chain(self.defer_events, self.io_writer_events, self.time_events):
            event.free()
        self.retval = retval


class PythonIOEvent:
    __slots__ = ('python_main_loop', 'fd', 'callback', 'userdata', 'on_destroy_callback', 'writer', 'reader',
                 'self_pointer')

    def __init__(self, python_main_loop: PythonMainLoop, fd: int, callback: pa_io_event_cb_t,
                 userdata: c.c_void_p) -> None:
        self.python_main_loop = python_main_loop
        self.fd = fd
        self.callback = callback
        self.userdata = userdata
        self.on_destroy_callback: Optional[pa_io_event_destroy_cb_t] = None
        self.writer = False
        self.reader = False
        self.self_pointer: pa_io_event_p = c.cast(c.pointer(c.py_object(self)), pa_io_event_p)
        python_main_loop.io_events.add(self)

    def read(self) -> None:
        self.callback(self.python_main_loop.api_pointer, self.self_pointer.value, self.fd, PA_IO_EVENT_INPUT,
                      self.userdata)

    def write(self) -> None:
        self.callback(self.python_main_loop.api_pointer, self.self_pointer.value, self.fd, PA_IO_EVENT_OUTPUT,
                      self.userdata)

    def free(self) -> None:
        self.python_main_loop.register_unregister_io_event(self, False, False)
        if self.on_destroy_callback is not None:
            self.on_destroy_callback(self.python_main_loop.api_pointer, c.pointer(c.py_object(self)), self.userdata)
        self.python_main_loop.io_events.discard(self)

    def set_destroy(self, callback: pa_io_event_destroy_cb_t) -> None:
        self.on_destroy_callback = callback


class PythonTimeEvent:
    __slots__ = ('python_main_loop', 'callback', 'userdata', 'on_destroy_callback', 'handle', 'self_pointer')

    def __init__(self, python_main_loop: PythonMainLoop, callback: pa_time_event_cb_t, userdata: c.c_void_p) -> None:
        self.python_main_loop = python_main_loop
        self.callback = callback
        self.userdata = userdata
        self.on_destroy_callback: Optional[pa_io_event_destroy_cb_t] = None
        self.handle: Optional[asyncio.TimerHandle] = None
        self.self_pointer: pa_io_event_p = c.cast(c.pointer(c.py_object(self)), pa_io_event_p)
        python_main_loop.time_events.add(self)

    def restart(self, ts: timeval) -> None:
        if self.handle is not None:
            self.handle.cancel()
        self.handle = self.python_main_loop.loop.call_later(
            ts.to_float() - time.time(), self.callback, self.python_main_loop.api_pointer, self.self_pointer.value,
            ts, self.userdata)

    def free(self) -> None:
        if self.on_destroy_callback is not None:
            self.on_destroy_callback(self.python_main_loop.api_pointer, c.pointer(c.py_object(self)), self.userdata)
        if self.handle:
            self.handle.cancel()
        self.python_main_loop.time_events.discard(self)

    def set_destroy(self, callback: pa_io_event_destroy_cb_t) -> None:
        self.on_destroy_callback = callback


class PythonDeferEvent:
    __slots__ = ('python_main_loop', 'callback', 'userdata', 'on_destroy_callback', 'enabled', 'self_pointer', 'handle')

    def __init__(self, python_main_loop: PythonMainLoop, callback: pa_defer_event_cb_t, userdata: c.c_void_p) -> None:
        self.python_main_loop = python_main_loop
        self.callback = callback
        self.userdata = userdata
        self.on_destroy_callback: Optional[pa_io_event_destroy_cb_t] = None
        self.enabled = False
        self.self_pointer: pa_defer_event_p = c.cast(c.pointer(c.py_object(self)), pa_defer_event_p)
        python_main_loop.defer_events.add(self)
        self.handle = None

    def call(self) -> None:
        self.handle = None
        self.callback(self.python_main_loop.api_pointer, self.self_pointer.value, self.userdata)
        if self.enabled and self.handle is None:
            self.handle = self.python_main_loop.loop.call_soon(self.call)

    def enable(self, enable: bool) -> None:
        if enable and self.handle is None:
            self.handle = self.python_main_loop.loop.call_soon(self.call)
        elif not enable and self.handle is not None:
            self.handle.cancel()
            self.handle = None
        self.enabled = enable

    def free(self) -> None:
        self.enable(False)
        if self.on_destroy_callback is not None:
            self.on_destroy_callback(self.python_main_loop.api_pointer, c.pointer(c.py_object(self)), self.userdata)
        self.python_main_loop.defer_events.discard(self)

    def set_destroy(self, callback: pa_io_event_destroy_cb_t) -> None:
        self.on_destroy_callback = callback


@pa_io_new_t
def aio_io_new(main_loop: c.POINTER(pa_mainloop_api), fd: int, flags: int, cb: pa_io_event_cb_t,
               userdata: c.c_void_p) -> int:
    python_main_loop: PythonMainLoop = c.cast(main_loop.contents.userdata, c.POINTER(c.py_object)).contents.value

    event = PythonIOEvent(python_main_loop, fd, cb, userdata)

    reader = bool(flags & PA_IO_EVENT_INPUT)
    writer = bool(flags & PA_IO_EVENT_OUTPUT)
    python_main_loop.register_unregister_io_event(event, reader, writer)
    return c.cast(event.self_pointer, pa_io_event_p).value


@pa_io_enable_t
def aio_io_enable(e: pa_io_event_p, flags: int) -> None:
    event: PythonIOEvent = c.cast(e, c.POINTER(c.py_object)).contents.value

    reader = bool(flags & PA_IO_EVENT_INPUT)
    writer = bool(flags & PA_IO_EVENT_OUTPUT)
    event.python_main_loop.register_unregister_io_event(event, reader, writer)


@pa_io_set_destroy_t
def aio_io_set_destroy(e: pa_io_event_p, cb: pa_io_event_destroy_cb_t) -> None:
    event: PythonIOEvent = c.cast(e, c.POINTER(c.py_object)).contents.value
    event.set_destroy(cb)


@pa_io_free_t
def aio_io_free(e: pa_io_event_p) -> None:
    event: PythonIOEvent = c.cast(e, c.POINTER(c.py_object)).contents.value
    event.free()


@pa_time_new_t
def aio_time_new(main_loop: c.POINTER(pa_mainloop_api), ts: c.POINTER(timeval), cb: pa_io_event_cb_t,
                 userdata: c.c_void_p) -> int:
    python_main_loop: PythonMainLoop = c.cast(main_loop.contents.userdata, c.POINTER(c.py_object)).contents.value
    event = PythonTimeEvent(python_main_loop, cb, userdata)
    event.restart(ts.contents)
    return c.cast(event.self_pointer, pa_time_event_p).value


@pa_time_restart_t
def aio_time_restart(e: pa_time_event_p, ts: c.POINTER(timeval)) -> None:
    event: PythonTimeEvent = c.cast(e, c.POINTER(c.py_object)).contents.value
    event.restart(ts.contents)


@pa_time_set_destroy_t
def aio_time_set_destroy(e: pa_time_event_p, cb: pa_time_event_destroy_cb_t) -> None:
    event: PythonTimeEvent = c.cast(e, c.POINTER(c.py_object)).contents.value
    event.set_destroy(cb)


@pa_time_free_t
def aio_time_free(e: pa_io_event_p) -> None:
    event: PythonTimeEvent = c.cast(e, c.POINTER(c.py_object)).contents.value
    event.free()


@pa_defer_new_t
def aio_defer_new(main_loop: c.POINTER(pa_mainloop_api), cb: pa_defer_event_cb_t, userdata: c.c_void_p) -> int:
    python_main_loop: PythonMainLoop = c.cast(main_loop.contents.userdata, c.POINTER(c.py_object)).contents.value
    event = PythonDeferEvent(python_main_loop, cb, userdata)
    event.enable(True)
    return event.self_pointer.value


@pa_defer_enable_t
def aio_defer_enable(e: pa_defer_event_p, enable: bool) -> None:
    event: PythonDeferEvent = c.cast(e, c.POINTER(c.py_object)).contents.value
    event.enable(enable)


@pa_defer_set_destroy_t
def aio_defer_set_destroy(e: pa_defer_event_p, cb: pa_defer_event_destroy_cb_t) -> None:
    event: PythonDeferEvent = c.cast(e, c.POINTER(c.py_object)).contents.value
    event.set_destroy(cb)


@pa_defer_free_t
def aio_defer_free(e: pa_io_event_p) -> None:
    event: PythonDeferEvent = c.cast(e, c.POINTER(c.py_object)).contents.value
    event.free()


@pa_quit_t
def aio_quit(main_loop: c.POINTER(pa_mainloop_api), retval: int) -> None:
    python_main_loop: PythonMainLoop = c.cast(main_loop.contents.userdata, c.POINTER(c.py_object)).contents.value
    python_main_loop.stop(retval)



"""
This module defines tests for the asynchronous high-level Python interface to libpulse.

The code of this module is completely based on *pulsectl* version 20.5.1 (Git revision 471428c) and has been modified
to test the asynchronous interface of the `PulseAsync` class instead of the `Pulse` class of *pulsectl*. To avoid
redundant code, the init and cleanup functions for setting up the dummy PulseAudio server instance are imported from
*pulsectl*'s tests module, rather than redefining them here.

Copyright (c) 2016 Mike Kazantsev, 2021 Michael Thies
"""
import asyncio
import atexit
import functools
import os
import signal
import sys
import unittest

import pulsectl
import pulsectl_asyncio
from pulsectl.pulsectl import unicode
from pulsectl.tests.test_with_dummy_instance import dummy_pulse_init, dummy_pulse_cleanup


def async_test(f):
	"""
	Decorator to transform async unittest coroutines into normal test methods
	"""
	@functools.wraps(f)
	def wrapper(*args, **kwargs):
		loop = asyncio.get_event_loop()
		loop.run_until_complete(f(*args, **kwargs))
	return wrapper


class AsyncDummyTests(unittest.TestCase):

	proc = tmp_dir = None

	@classmethod
	def setUpClass(cls):
		assert not cls.proc and not cls.tmp_dir, [cls.proc, cls.tmp_dir]

		for sig in 'hup', 'term', 'int':
			signal.signal(getattr(signal, 'sig{}'.format(sig).upper()), lambda sig,frm: sys.exit())
		atexit.register(cls.tearDownClass)

		cls.instance_info = dummy_pulse_init()
		for k, v in cls.instance_info.items(): setattr(cls, k, v)

	@classmethod
	def tearDownClass(cls):
		dummy_pulse_cleanup(cls.instance_info)
		cls.proc = cls.tmp_dir = None


	# Fuzzy float comparison is necessary for volume,
	#  as these loose precision when converted to/from pulse int values.

	_compare_floats_rounding = 3
	def _compare_floats(self, a, b, msg=None):
		if round(a, self._compare_floats_rounding) != round(b, self._compare_floats_rounding):
			return self._baseAssertEqual(a, b, msg)

	def __init__(self, *args, **kws):
		super(AsyncDummyTests, self).__init__(*args, **kws)
		self.addTypeEqualityFunc(float, self._compare_floats)


	@async_test
	async def test_connect(self):
		with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			si = await pulse.server_info()
		with pulsectl_asyncio.PulseAsync('t', server=self.sock_tcp4) as pulse:
			await pulse.connect()
			si4 = await pulse.server_info()
		self.assertEqual(vars(si), vars(si4))
		if int(self.sock_tcp6.rsplit(':', 1)[-1]):
			with pulsectl_asyncio.PulseAsync('t', server=self.sock_tcp6) as pulse:
				await pulse.connect()
				si6 = await pulse.server_info()
			self.assertEqual(vars(si), vars(si6))

	@async_test
	async def test_connect_timeout(self):
		self.sock_delay_thread_ready.wait(timeout=2)
		async with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			si = await pulse.server_info()

		async with pulsectl_asyncio.PulseAsync('t', server=self.sock_tcp_delay) as pulse:
			sid = await pulse.server_info()
		self.assertEqual(vars(si), vars(sid))
		self.sock_delay_thread_disco.set()

		with pulsectl_asyncio.PulseAsync('t', server=self.sock_tcp_delay) as pulse:
			await pulse.connect()
			sid = await pulse.server_info()
		self.assertEqual(vars(si), vars(sid))
		self.sock_delay_thread_disco.set()

		with pulsectl_asyncio.PulseAsync('t', server=self.sock_tcp_delay) as pulse:
			await pulse.connect(timeout=1.0)
			sid = await pulse.server_info()
		self.assertEqual(vars(si), vars(sid))
		self.sock_delay_thread_disco.set()

		with pulsectl_asyncio.PulseAsync('t', server=self.sock_tcp_delay) as pulse:
			with self.assertRaises(asyncio.TimeoutError):
				await pulse.connect(timeout=0.1)
			self.sock_delay_thread_disco.set()
			await pulse.connect(timeout=1.0)
			sid = await pulse.server_info()
		self.assertEqual(vars(si), vars(sid))
		self.sock_delay_thread_disco.set()

	@async_test
	async def test_server_info(self):
		with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			si, srcs, sinks = await pulse.server_info(), await pulse.source_list(), await pulse.sink_list()
		self.assertEqual(len(srcs), 2)
		self.assertEqual(len(sinks), 2)

	@async_test
	async def test_default_set(self):
		with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			(src1, src2), (sink1, sink2) = (await pulse.source_list())[:2], (await pulse.sink_list())[:2]
			self.assertNotEqual(sink1.name, sink2.name)
			self.assertNotEqual(src1.name, src2.name)

			await pulse.default_set(sink1)
			await pulse.default_set(sink1)
			await pulse.default_set(src1)
			si = await pulse.server_info()
			self.assertEqual(si.default_sink_name, sink1.name)
			self.assertEqual(si.default_source_name, src1.name)

			await pulse.default_set(sink2)
			si = await pulse.server_info()
			self.assertEqual(si.default_sink_name, sink2.name)
			self.assertEqual(si.default_source_name, src1.name)

			await pulse.default_set(src2)
			await pulse.default_set(src2)
			await pulse.default_set(sink1)
			si = await pulse.server_info()
			self.assertEqual(si.default_sink_name, sink1.name)
			self.assertEqual(si.default_source_name, src2.name)

			await pulse.sink_default_set(sink2.name)
			await pulse.source_default_set(src1.name)
			si = await pulse.server_info()
			self.assertEqual(si.default_sink_name, sink2.name)
			self.assertEqual(si.default_source_name, src1.name)

			nx = 'xxx'
			self.assertNotIn(nx, [sink1.name, sink2.name])
			self.assertNotIn(nx, [src1.name, src2.name])
			with self.assertRaises(TypeError): await pulse.sink_default_set(sink2.index)
			with self.assertRaises(pulsectl.PulseOperationFailed): await pulse.sink_default_set(nx)
			with self.assertRaises(pulsectl.PulseOperationFailed): await pulse.source_default_set(nx)
			si = await pulse.server_info()
			self.assertEqual(si.default_sink_name, sink2.name)
			self.assertEqual(si.default_source_name, src1.name)

	@async_test
	async def test_events(self):
		with pulsectl_asyncio.PulseAsync ('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			sink, cb_called = (await pulse.sink_list())[0], list()

			async def listen_events():
				async for ev in pulse.subscribe_events('all'):
					self.assertEqual(ev.facility, 'sink')
					self.assertEqual(ev.t, 'change')
					self.assertEqual(ev.index, sink.index)
					cb_called.append(True)
					break
			loop = asyncio.get_event_loop()
			loop.create_task(listen_events())
			await asyncio.sleep(0)

			await pulse.volume_set_all_chans(sink, 0.6)
			await asyncio.sleep(0.05)
			self.assertTrue(bool(cb_called))
			self.assertIsNone(pulse.event_callback)

	@async_test
	async def test_sink_src(self):
		with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			src, sink = (await pulse.source_list())[0], (await pulse.sink_list())[0]
			self.assertTrue(src.proplist.get('device.class'))
			self.assertTrue(isinstance(src.proplist.get('device.class'), unicode))
			self.assertTrue(isinstance(list(src.proplist.keys())[0], unicode))
			self.assertTrue(sink.proplist.get('device.class'))
			self.assertTrue(isinstance(sink.proplist.get('device.class'), unicode))
			self.assertTrue(isinstance(list(sink.proplist.keys())[0], unicode))

			await pulse.mute(src, False)
			print(repr(src))
			self.assertFalse(src.mute)
			self.assertFalse((await pulse.source_info(src.index)).mute)
			await pulse.mute(src, True)
			await pulse.mute(src, True)
			self.assertTrue(src.mute)
			self.assertTrue((await pulse.source_info(src.index)).mute)
			await pulse.mute(src, False)

			await pulse.mute(sink, False)
			self.assertFalse(sink.mute)
			self.assertFalse((await pulse.sink_info(sink.index)).mute)
			await pulse.mute(sink)
			self.assertTrue(sink.mute)
			self.assertTrue((await pulse.sink_info(sink.index)).mute)
			await pulse.mute(sink, False)

			await pulse.volume_set_all_chans(sink, 1.0)
			self.assertEqual(sink.volume.value_flat, 1.0)
			self.assertEqual((await pulse.sink_info(sink.index)).volume.values, sink.volume.values)
			await pulse.volume_set_all_chans(sink, 0.5)
			self.assertEqual(sink.volume.value_flat, 0.5)
			self.assertEqual((await pulse.sink_info(sink.index)).volume.values, sink.volume.values)
			await pulse.volume_change_all_chans(sink, -0.5)
			self.assertEqual(sink.volume.value_flat, 0.0)
			self.assertEqual((await pulse.sink_info(sink.index)).volume.values, sink.volume.values)
			await pulse.volume_set_all_chans(sink, 1.0)

	@async_test
	async def test_get_sink_src(self):
		with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			src, sink = await pulse.source_list(), await pulse.sink_list()
			src_nx, sink_nx = max(s.index for s in src)+1, max(s.index for s in sink)+1
			src, sink = src[0], sink[0]
			self.assertEqual(sink.index, (await pulse.get_sink_by_name(sink.name)).index)
			self.assertEqual(src.index, (await pulse.get_source_by_name(src.name)).index)
			with self.assertRaises(pulsectl.PulseIndexError): await pulse.source_info(src_nx)
			with self.assertRaises(pulsectl.PulseIndexError): await pulse.sink_info(sink_nx)

	# def test_get_card(self): no cards to test these calls with :(

	@async_test
	async def test_module_funcs(self):
		with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			self.assertEqual(len(await pulse.sink_list()), 2)
			idx = await pulse.module_load('module-null-sink')
			self.assertEqual(len(await pulse.sink_list()), 3)
			await pulse.module_unload(idx)
			self.assertEqual(len(await pulse.sink_list()), 2)
			with self.assertRaises(pulsectl.PulseError):
				await pulse.module_load('module-that-does-not-exist')
			self.assertEqual(len(await pulse.sink_list()), 2)

	@async_test
	async def test_stream(self):
		with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			stream_started = asyncio.Event()
			stream_idx = []

			async def listen_stream_events():
				async for ev in pulse.subscribe_events('sink_input'):
					if ev.t == 'new':
						stream_idx.append(ev.index)
						stream_started.set()
						break

			loop = asyncio.get_event_loop()
			loop.create_task(listen_stream_events())

			paplay = await asyncio.create_subprocess_exec(
				'paplay', '--raw', '/dev/zero', env=dict(
					PATH=os.environ['PATH'], XDG_RUNTIME_DIR=self.tmp_dir))
			try:
				await stream_started.wait()
				self.assertTrue(bool(stream_idx))
				stream_idx = stream_idx[0]

				stream = await pulse.sink_input_info(stream_idx)
				self.assertTrue(stream.proplist.get('application.name'))
				self.assertTrue(isinstance(stream.proplist.get('application.name'), unicode))
				self.assertTrue(isinstance(list(stream.proplist.keys())[0], unicode))

				await pulse.mute(stream, False)
				self.assertFalse(stream.mute)
				self.assertFalse((await pulse.sink_input_info(stream.index)).mute)
				await pulse.mute(stream)
				self.assertTrue(stream.mute)
				self.assertTrue((await pulse.sink_input_info(stream.index)).mute)
				await pulse.mute(stream, False)

				await pulse.volume_set_all_chans(stream, 1.0)
				self.assertEqual(stream.volume.value_flat, 1.0)
				self.assertEqual((await pulse.sink_input_info(stream.index)).volume.values, stream.volume.values)
				await pulse.volume_set_all_chans(stream, 0.5)
				self.assertEqual(stream.volume.value_flat, 0.5)
				self.assertEqual((await pulse.sink_input_info(stream.index)).volume.values, stream.volume.values)
				await pulse.volume_change_all_chans(stream, -0.5)
				self.assertEqual(stream.volume.value_flat, 0.0)
				self.assertEqual((await pulse.sink_input_info(stream.index)).volume.values, stream.volume.values)

			finally:
				if paplay.returncode is None:
					paplay.kill()
				await paplay.wait()

			with self.assertRaises(pulsectl.PulseIndexError):
				await pulse.sink_input_info(stream.index)

	@async_test
	async def test_ext_stream_restore(self):
		sr_name1 = 'sink-input-by-application-name:pulsectl-test-1'
		sr_name2 = 'sink-input-by-application-name:pulsectl-test-2'

		with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			self.assertIsNotNone(await pulse.stream_restore_test())

			await pulse.stream_restore_write(sr_name1, volume=0.5, mute=True)
			await pulse.stream_restore_write(
				pulsectl.PulseExtStreamRestoreInfo(sr_name2, volume=0.3, channel_list='mono'),
				apply_immediately=True )

			sr_list = await pulse.stream_restore_list()
			self.assertIsInstance(sr_list, list)
			self.assertTrue(sr_list)
			sr_dict = dict((sr.name, sr) for sr in sr_list)
			self.assertEqual(sr_dict[sr_name1].volume.value_flat, 0.5)
			self.assertEqual(sr_dict[sr_name1].mute, 1)
			self.assertEqual(sr_dict[sr_name1].channel_list, [pulse.channel_list_enum.mono])
			self.assertIn(sr_name2, sr_dict)
			self.assertEqual(sr_dict[sr_name1].channel_list, [pulse.channel_list_enum.mono])
			self.assertEqual(sr_dict[sr_name1].channel_list_raw, [0])

			await pulse.stream_restore_delete(sr_name1)
			sr_dict = dict((sr.name, sr) for sr in await pulse.stream_restore_list())
			self.assertNotIn(sr_name1, sr_dict)
			self.assertIn(sr_name2, sr_dict)

			await pulse.stream_restore_write(
				[ pulsectl.PulseExtStreamRestoreInfo( sr_name1,
						volume=0.7, channel_list=['front-left', 'front-right'] ),
					sr_dict[sr_name2] ],
				mode='merge' )
			await pulse.stream_restore_write(sr_name1,
				volume=0.3, channel_list='mono', mute=True )
			sr_dict = dict((sr.name, sr) for sr in await pulse.stream_restore_list())
			self.assertEqual(sr_dict[sr_name1].volume.value_flat, 0.7)
			self.assertEqual(sr_dict[sr_name1].mute, 0)
			self.assertEqual(sr_dict[sr_name1].channel_list,
				[pulse.channel_list_enum.front_left, pulse.channel_list_enum.front_right])
			self.assertEqual(sr_dict[sr_name1].channel_list_raw, [1, 2])

			await pulse.stream_restore_write(sr_name1, volume=0.4, mode='replace')
			sr_dict = dict((sr.name, sr) for sr in await pulse.stream_restore_list())
			self.assertEqual(sr_dict[sr_name1].volume.value_flat, 0.4)

			await pulse.stream_restore_write(sr_name2, volume=0.9, mode='set')
			sr_dict = dict((sr.name, sr) for sr in await pulse.stream_restore_list())
			self.assertEqual(sr_dict[sr_name2].volume.value_flat, 0.9)
			self.assertEqual(list(sr_dict.keys()), [sr_name2])

			await pulse.stream_restore_write([], mode='set') # i.e. remove all
			sr_dict = dict((sr.name, sr) for sr in await pulse.stream_restore_list())
			self.assertNotIn(sr_name1, sr_dict)
			self.assertNotIn(sr_name2, sr_dict)

	@async_test
	async def test_stream_move(self):
		with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			stream_started = asyncio.Event()
			stream_idx = []

			async def listen_stream_events():
				async for ev in pulse.subscribe_events('sink_input'):
					if ev.t == 'new':
						stream_idx.append(ev.index)
						stream_started.set()
						break

			loop = asyncio.get_event_loop()
			loop.create_task(listen_stream_events())

			paplay = await asyncio.create_subprocess_exec(
				'paplay', '--raw', '/dev/zero', env=dict(
					PATH=os.environ['PATH'], XDG_RUNTIME_DIR=self.tmp_dir))
			try:
				await stream_started.wait()
				self.assertTrue(bool(stream_idx))
				stream_idx = stream_idx[0]

				stream = await pulse.sink_input_info(stream_idx)
				sink_indexes = set(s.index for s in await pulse.sink_list())
				sink1 = stream.sink
				sink2 = sink_indexes.difference([sink1]).pop()
				sink_nx = max(sink_indexes) + 1

				await pulse.sink_input_move(stream.index, sink2)
				stream_new = await pulse.sink_input_info(stream.index)
				self.assertEqual(stream.sink, sink1) # old info doesn't get updated
				self.assertEqual(stream_new.sink, sink2)

				await pulse.sink_input_move(stream.index, sink1) # move it back
				stream_new = await pulse.sink_input_info(stream.index)
				self.assertEqual(stream_new.sink, sink1)

				with self.assertRaises(pulsectl.PulseOperationFailed):
					await pulse.sink_input_move(stream.index, sink_nx)

			finally:
				if paplay.returncode is None:
					paplay.kill()
				await paplay.wait()

	@async_test
	async def test_get_peak_sample(self):
		# Note: this test takes at least multiple seconds to run
		with pulsectl_asyncio.PulseAsync('t', server=self.sock_unix) as pulse:
			await pulse.connect()
			source_any = max(s.index for s in await pulse.source_list())
			source_nx = source_any + 1

			await asyncio.sleep(0.3) # make sure previous streams die
			peak = await pulse.get_peak_sample(source_any, 0.3)
			self.assertEqual(peak, 0)

			stream_started = asyncio.Event()
			stream_idx = []

			async def listen_stream_events():
				async for ev in pulse.subscribe_events('sink_input'):
					if ev.t == 'new':
						stream_idx.append(ev.index)
						stream_started.set()
						break

			loop = asyncio.get_event_loop()
			loop.create_task(listen_stream_events())

			paplay = await asyncio.create_subprocess_exec(
				'paplay', '--raw', '/dev/urandom', env=dict(
					PATH=os.environ['PATH'], XDG_RUNTIME_DIR=self.tmp_dir))
			try:
				await stream_started.wait()
				self.assertTrue(bool(stream_idx))
				stream_idx = stream_idx[0]
				si = await pulse.sink_input_info(stream_idx)
				sink = await pulse.sink_info(si.sink)
				source = await pulse.source_info(sink.monitor_source)

				# First poll can randomly fail if too short, probably due to latency or such
				peak = await pulse.get_peak_sample(sink.monitor_source, 3)
				self.assertGreater(peak, 0)

				peak = await pulse.get_peak_sample(source.index, 0.3, si.index)
				self.assertGreater(peak, 0)
				peak = await pulse.get_peak_sample(source.name, 0.3, si.index)
				self.assertGreater(peak, 0)
				peak = await pulse.get_peak_sample(source_nx, 0.3)
				self.assertEqual(peak, 0)

				paplay.terminate()
				await paplay.wait()

				peak = await pulse.get_peak_sample(source.index, 0.3, si.index)
				self.assertEqual(peak, 0)

			finally:
				if paplay.returncode is None:
					paplay.kill()
				await paplay.wait()


class PulseCrashTestsAsync(unittest.TestCase):

	@classmethod
	def setUpClass(cls):
		for sig in 'hup', 'term', 'int':
			signal.signal(getattr(signal, 'sig{}'.format(sig).upper()), lambda sig,frm: sys.exit())

	@async_test
	async def test_crash_after_connect(self):
		loop = asyncio.get_event_loop()
		info = await loop.run_in_executor(None, dummy_pulse_init)
		try:
			with pulsectl_asyncio.PulseAsync('t', server=info.sock_unix) as pulse:
				await pulse.connect()
				for si in await pulse.sink_list(): self.assertTrue(si)
				await loop.run_in_executor(None, info.proc.terminate)
				await loop.run_in_executor(None, info.proc.wait)
				with self.assertRaises((pulsectl._pulsectl.LibPulse.CallError, pulsectl.PulseOperationFailed)):
					for si in await pulse.sink_list(): raise AssertionError(si)
				self.assertFalse(pulse.connected)
		finally: await loop.run_in_executor(None, dummy_pulse_cleanup, info)

	@async_test
	async def test_reconnect(self):
		loop = asyncio.get_event_loop()
		info = await loop.run_in_executor(None, dummy_pulse_init)
		try:
			with pulsectl_asyncio.PulseAsync('t', server=info.sock_unix) as pulse:
				with self.assertRaises(Exception):
					for si in await pulse.sink_list(): raise AssertionError(si)

				await pulse.connect(autospawn=False)
				self.assertTrue(pulse.connected)
				for si in await pulse.sink_list(): self.assertTrue(si)
				await loop.run_in_executor(None, info.proc.terminate)
				await loop.run_in_executor(None, info.proc.wait)
				with self.assertRaises(Exception):
					for si in await pulse.sink_list(): raise AssertionError(si)
				self.assertFalse(pulse.connected)

				await loop.run_in_executor(None, dummy_pulse_init, info)
				await pulse.connect(autospawn=False, wait=True)
				self.assertTrue(pulse.connected)
				for si in await pulse.sink_list(): self.assertTrue(si)

				pulse.disconnect()
				with self.assertRaises(Exception):
					for si in await pulse.sink_list(): raise AssertionError(si)
				self.assertFalse(pulse.connected)
				await pulse.connect(autospawn=False)
				self.assertTrue(pulse.connected)
				for si in await pulse.sink_list(): self.assertTrue(si)

		finally: await loop.run_in_executor(None, dummy_pulse_cleanup, info)



import atexit
import os
import signal
import sys
import unittest
import io

from pulsectl.tests.test_with_dummy_instance import dummy_pulse_cleanup, dummy_pulse_init


class ExamplesTest(unittest.TestCase):
	proc = tmp_dir = None

	@classmethod
	def setUpClass(cls):
		assert not cls.proc and not cls.tmp_dir, [cls.proc, cls.tmp_dir]
		for sig in 'hup', 'term', 'int':
			signal.signal(getattr(signal, 'sig{}'.format(sig).upper()), lambda sig,frm: sys.exit())
		atexit.register(cls.tearDownClass)
		cls.instance_info = dummy_pulse_init()
		for k, v in cls.instance_info.items():
			setattr(cls, k, v)
		os.environ['XDG_RUNTIME_DIR'] = cls.tmp_dir

	@classmethod
	def tearDownClass(cls):
		dummy_pulse_cleanup(cls.instance_info)
		cls.proc = cls.tmp_dir = None

	def test_simple(self):
		import examples.simple_example

	def test_subscribe(self):
		import examples.subscribe_example

	def test_subscribe_peak(self):
		_stdout = sys.stdout
		sys.stdout = io.StringIO()
		try:
			import examples.subscribe_peak_example
		finally:
			sys.stdout = _stdout



"""
This module provides an asynchronous high-level interface for PulseAudio actions, based on ctypes wrappers of libpulse
and PulseAudio object abstraction from *pulsectl* package.

The code of this module is completely based on *pulsectl* version 20.5.1 (Git revision 471428c). PulseObject class
definitions have been removed to be imported from the original *pulsectl* module instead. The `Pulse` class has been
modified to the `PulseAsync` class with an asynchronous interface.

Copyright (c) 2014 George Filipkin, 2016 Mike Kazantsev, 2021 Michael Thies
"""
import asyncio
import inspect
import functools as ft
from typing import Optional, AsyncIterator, Coroutine
from contextlib import suppress
import sys
from warnings import warn

from .pa_asyncio_mainloop import PythonMainLoop
from pulsectl.pulsectl import (
	PulseError, PulseEventTypeEnum, PulseEventFacilityEnum, PulseEventInfo,
	PulseEventMaskEnum, PulseLoopStop, PulseOperationFailed, PulseIndexError, PulseSinkInfo, PulseSourceInfo,
	PulseCardInfo, PulseSinkInputInfo, PulseSourceOutputInfo, PulseClientInfo, PulseServerInfo, PulseModuleInfo,
	is_list, PulseOperationInvalid, PulsePortInfo, PulseExtStreamRestoreInfo, PulseUpdateEnum, is_str,
	assert_pulse_object, PulseDisconnected, unicode, Enum)
from pulsectl import _pulsectl as c


class _pulse_op_cb:
	"""
	Async context manager class, used to create a future in PulseAsync's _actions dict and the corresponding
	callback method to be passed to the PulseAudio action call for resolving the future. The `__aexit__()` method is
	the place, where the actual asynchronous magic of this library happens.

	In `pulsectl`, it's implemented as a method of the Pulse class with @contextmanager, however, we need
	@asynccontextmanager here, which is only available since Python 3.7.
	"""
	def __init__(self, async_pulse: "PulseAsync", raw=False):
		self.raw = raw
		self.future = None
		self.async_pulse = async_pulse

	async def __aenter__(self):
		loop = asyncio.get_event_loop()
		self.future = loop.create_future()

		def cb(s=True):
			if s:
				loop.call_soon_threadsafe(self.future.set_result, None)
			else:
				loop.call_soon_threadsafe(self.future.set_exception, PulseOperationFailed())

		if not self.raw:
			cb = c.PA_CONTEXT_SUCCESS_CB_T(lambda ctx, s, d, cb=cb: cb(s))
		self.async_pulse.waiting_futures.add(self.future)
		return cb

	async def __aexit__(self, exc_type, exc_val, exc_tb):
		try:
			if not exc_type:
				await self.future
		finally:
			self.async_pulse.waiting_futures.discard(self.future)


class PulseAsync(object):

	_ctx = None

	def __init__(self, client_name=None, server=None, loop: Optional[asyncio.AbstractEventLoop] = None):
		'''Connects to specified pulse server by default.
			Specifying "connect=False" here prevents that, but be sure to call connect() later.
			"connect=False" can also be used here to
				have control over options passed to connect() method.'''
		if loop:
			warn("The parameter 'loop' is deprecated and will be removed in future versions, since it is no longer necessary", DeprecationWarning)

		self.name = client_name or 'pulsectl'
		self.server = server
		self._connected = asyncio.Event(**({"loop": loop} if sys.version_info[:2] < (3, 8) else {}))
		self._disconnected = asyncio.Event(**({"loop": loop} if sys.version_info[:2] < (3, 8) else {}))
		self._disconnected.set()
		self._ctx = self._loop = None
		self.init(loop)

	def init(self, loop: Optional[asyncio.AbstractEventLoop]):
		self._pa_state_cb = c.PA_STATE_CB_T(self._pulse_state_cb)
		self._pa_subscribe_cb = c.PA_SUBSCRIBE_CB_T(self._pulse_subscribe_cb)

		self._loop = PythonMainLoop(loop or asyncio.get_event_loop())

		self._ctx_init()
		self.event_types = sorted(PulseEventTypeEnum._values.values())
		self.event_facilities = sorted(PulseEventFacilityEnum._values.values())
		self.event_masks = sorted(PulseEventMaskEnum._values.values())
		self.event_callback = None
		self.waiting_futures = set()

		chan_names = dict()
		for n in range(256):
			name = c.pa.channel_position_to_string(n)
			if name is None: break
			chan_names[n] = name
		self.channel_list_enum = Enum('channel_pos', chan_names)

	def _ctx_init(self):
		if self._ctx:
			self.disconnect()
			c.pa.context_unref(self._ctx)
		self._ctx = c.pa.context_new(self._loop.api_pointer, self.name)
		self._connected.clear()
		self._disconnected.clear()
		c.pa.context_set_state_callback(self._ctx, self._pa_state_cb, None)
		c.pa.context_set_subscribe_callback(self._ctx, self._pa_subscribe_cb, None)

	async def connect(self, autospawn=False, wait=False, timeout=None):
		'''Connect to pulseaudio server.
			"autospawn" option will start new pulse daemon, if necessary.
			Specifying "wait" option will make function block until pulseaudio server appears.
			"timeout" (in seconds) will raise asyncio.TimeoutError if connection not established within it.'''
		if self._connected.is_set() or self._disconnected.is_set():
			self._ctx_init()
		flags = 0
		if not autospawn:
			flags |= c.PA_CONTEXT_NOAUTOSPAWN
		if wait:
			flags |= c.PA_CONTEXT_NOFAIL
		try:
			c.pa.context_connect(self._ctx, self.server, flags, None)
			if not timeout:
				await self._wait_disconnect_or(self._connected.wait())
			else:
				await asyncio.wait_for(self._wait_disconnect_or(self._connected.wait()), timeout)
		except (c.pa.CallError, PulseDisconnected) as e:
			self._disconnected.set()
			raise PulseError('Failed to connect to pulseaudio server') from e
		except asyncio.TimeoutError:
			self.disconnect()
			await self._disconnected.wait()
			raise

	@property
	def connected(self):
		return self._connected.is_set()

	def disconnect(self):
		if not self._ctx or self._disconnected.is_set():
			return
		c.pa.context_disconnect(self._ctx)

	def close(self):
		if not self._loop: return
		try:
			self.disconnect()
			c.pa.context_unref(self._ctx)
			self._loop.stop(0)
		finally: self._ctx = self._loop = None

	def __enter__(self): return self
	def __exit__(self, err_t, err, err_tb): self.close()

	async def __aenter__(self):
		await self.connect()
		return self

	async def __aexit__(self, err_t, err, err_tb):
		self.close()

	async def _wait_disconnect_or(self, coroutine: Coroutine):
		loop = asyncio.get_event_loop()
		wait_disconnected = loop.create_task(self._disconnected.wait())
		other_task = loop.create_task(coroutine)
		try:
			done, pending = await asyncio.wait((wait_disconnected, other_task), return_when=asyncio.FIRST_COMPLETED)
		except BaseException:  # Catches all Exception subclasses *and* (more important) CancelledError
			for task in (wait_disconnected, other_task):
				task.cancel()
			raise
		for task in pending:
			task.cancel()
		if other_task in pending:
			raise PulseDisconnected()
		else:
			return other_task.result()

	def _pulse_state_cb(self, ctx, _userdata):
		state = c.pa.context_get_state(ctx)
		if state >= c.PA_CONTEXT_READY:
			if state == c.PA_CONTEXT_READY:
				self._disconnected.clear()
				self._connected.set()
			elif state in [c.PA_CONTEXT_FAILED, c.PA_CONTEXT_TERMINATED]:
				self._connected.clear()
				self._disconnected.set()
				for future in self.waiting_futures:
					future.set_exception(PulseDisconnected())

	def _pulse_subscribe_cb(self, ctx, ev, idx, userdata):
		if not self.event_callback: return
		n = ev & c.PA_SUBSCRIPTION_EVENT_FACILITY_MASK
		ev_fac = PulseEventFacilityEnum._c_val(n, 'ev.facility.{}'.format(n))
		n = ev & c.PA_SUBSCRIPTION_EVENT_TYPE_MASK
		ev_t = PulseEventTypeEnum._c_val(n, 'ev.type.{}'.format(n))
		try: self.event_callback(PulseEventInfo(ev_t, ev_fac, idx))
		except PulseLoopStop: self._loop_stop = True

	def _pulse_info_cb(self, info_cls, data_list, done_cb, ctx, info, eof, userdata):
		# No idea where callbacks with "userdata != NULL" come from,
		#  but "info" pointer in them is always invalid, so they are discarded here.
		# Looks like some kind of mixup or corruption in libpulse memory?
		# See also: https://github.com/mk-fg/python-pulse-control/issues/35
		if userdata is not None: return
		# Empty result list and conn issues are checked elsewhere.
		# Errors here are non-descriptive (errno), so should not be useful anyway.
		# if eof < 0: done_cb(s=False)
		if eof: done_cb()
		else: data_list.append(info_cls(info[0]))

	def _pulse_get_list(cb_t, pulse_func, info_cls, singleton=False, index_arg=True):
		async def _wrapper_method(self, index=None):
			data = list()
			async with _pulse_op_cb(self, raw=True) as cb:
				cb = cb_t(
					ft.partial(self._pulse_info_cb, info_cls, data, cb) if not singleton else
					lambda ctx, info, userdata, cb=cb: data.append(info_cls(info[0])) or cb() )
				try:
					pa_op = pulse_func( self._ctx,
						*([index, cb, None] if index is not None else [cb, None]) )
				except c.ArgumentError as err: raise TypeError(err.args)
				except c.pa.CallError as err: raise PulseOperationInvalid(err.args[-1])
			c.pa.operation_unref(pa_op)
			data = data or list()
			if index is not None or singleton:
				if not data: raise PulseIndexError(index)
				data, = data
			return data
		_wrapper_method.__name__ = '...'
		_wrapper_method.__doc__ = 'Signature: func({})'.format(
			'' if pulse_func.__name__.endswith('_list') or singleton or not index_arg else 'index' )
		return _wrapper_method

	get_sink_by_name = _pulse_get_list(
		c.PA_SINK_INFO_CB_T,
		c.pa.context_get_sink_info_by_name, PulseSinkInfo )
	get_source_by_name = _pulse_get_list(
		c.PA_SOURCE_INFO_CB_T,
		c.pa.context_get_source_info_by_name, PulseSourceInfo )
	get_card_by_name = _pulse_get_list(
		c.PA_CARD_INFO_CB_T,
		c.pa.context_get_card_info_by_name, PulseCardInfo )

	sink_input_list = _pulse_get_list(
		c.PA_SINK_INPUT_INFO_CB_T,
		c.pa.context_get_sink_input_info_list, PulseSinkInputInfo )
	sink_input_info = _pulse_get_list(
		c.PA_SINK_INPUT_INFO_CB_T,
		c.pa.context_get_sink_input_info, PulseSinkInputInfo )
	source_output_list = _pulse_get_list(
		c.PA_SOURCE_OUTPUT_INFO_CB_T,
		c.pa.context_get_source_output_info_list, PulseSourceOutputInfo )
	source_output_info = _pulse_get_list(
		c.PA_SOURCE_OUTPUT_INFO_CB_T,
		c.pa.context_get_source_output_info, PulseSourceOutputInfo )

	sink_list = _pulse_get_list(
		c.PA_SINK_INFO_CB_T, c.pa.context_get_sink_info_list, PulseSinkInfo )
	sink_info = _pulse_get_list(
		c.PA_SINK_INFO_CB_T, c.pa.context_get_sink_info_by_index, PulseSinkInfo )
	source_list = _pulse_get_list(
		c.PA_SOURCE_INFO_CB_T, c.pa.context_get_source_info_list, PulseSourceInfo )
	source_info = _pulse_get_list(
		c.PA_SOURCE_INFO_CB_T, c.pa.context_get_source_info_by_index, PulseSourceInfo )
	card_list = _pulse_get_list(
		c.PA_CARD_INFO_CB_T, c.pa.context_get_card_info_list, PulseCardInfo )
	card_info = _pulse_get_list(
		c.PA_CARD_INFO_CB_T, c.pa.context_get_card_info_by_index, PulseCardInfo )
	client_list = _pulse_get_list(
		c.PA_CLIENT_INFO_CB_T, c.pa.context_get_client_info_list, PulseClientInfo )
	client_info = _pulse_get_list(
		c.PA_CLIENT_INFO_CB_T, c.pa.context_get_client_info, PulseClientInfo )
	server_info = _pulse_get_list(
		c.PA_SERVER_INFO_CB_T, c.pa.context_get_server_info, PulseServerInfo, singleton=True )
	module_info = _pulse_get_list(
		c.PA_MODULE_INFO_CB_T, c.pa.context_get_module_info, PulseModuleInfo )
	module_list = _pulse_get_list(
		c.PA_MODULE_INFO_CB_T, c.pa.context_get_module_info_list, PulseModuleInfo )

	def _pulse_method_call(pulse_op, func=None, index_arg=True):
		'''Creates following synchronous wrapper for async pa_operation callable:
			wrapper(index, ...) -> pulse_op(index, [*]args_func(...))
			index_arg=False: wrapper(...) -> pulse_op([*]args_func(...))'''
		async def _wrapper(self, *args, **kws):
			if index_arg:
				if 'index' in kws: index = kws.pop('index')
				else: index, args = args[0], args[1:]
			pulse_args = func(*args, **kws) if func else list()
			if not is_list(pulse_args): pulse_args = [pulse_args]
			if index_arg: pulse_args = [index] + list(pulse_args)
			async with _pulse_op_cb(self) as cb:
				try: pulse_op(self._ctx, *(list(pulse_args) + [cb, None]))
				except c.ArgumentError as err: raise TypeError(err.args)
				except c.pa.CallError as err: raise PulseOperationInvalid(err.args[-1])

		signature = inspect.signature(func or (lambda: None))
		if index_arg:
			signature.replace(
				parameters=
				[inspect.Parameter("index", inspect.Parameter.POSITIONAL_OR_KEYWORD)]
				+ list(signature.parameters.values()))
		_wrapper.__name__ = '...'
		_wrapper.__doc__ = 'Signature: func' + str(signature)
		if func.__doc__: _wrapper.__doc__ += '\n\n' + func.__doc__
		return _wrapper

	card_profile_set_by_index = _pulse_method_call(
		c.pa.context_set_card_profile_by_index, lambda profile_name: profile_name )

	sink_default_set = _pulse_method_call(
		c.pa.context_set_default_sink, index_arg=False,
		func=lambda sink: sink.name if isinstance(sink, PulseSinkInfo) else sink )
	source_default_set = _pulse_method_call(
		c.pa.context_set_default_source, index_arg=False,
		func=lambda source: source.name if isinstance(source, PulseSourceInfo) else source )

	sink_input_mute = _pulse_method_call(
		c.pa.context_set_sink_input_mute, lambda mute=True: mute )
	sink_input_move = _pulse_method_call(
		c.pa.context_move_sink_input_by_index, lambda sink_index: sink_index )
	sink_mute = _pulse_method_call(
		c.pa.context_set_sink_mute_by_index, lambda mute=True: mute )
	sink_input_volume_set = _pulse_method_call(
		c.pa.context_set_sink_input_volume, lambda vol: vol.to_struct() )
	sink_volume_set = _pulse_method_call(
		c.pa.context_set_sink_volume_by_index, lambda vol: vol.to_struct() )
	sink_suspend = _pulse_method_call(
		c.pa.context_suspend_sink_by_index, lambda suspend=True: suspend )
	sink_port_set = _pulse_method_call(
		c.pa.context_set_sink_port_by_index,
		lambda port: port.name if isinstance(port, PulsePortInfo) else port )

	source_output_mute = _pulse_method_call(
		c.pa.context_set_source_output_mute, lambda mute=True: mute )
	source_output_move = _pulse_method_call(
		c.pa.context_move_source_output_by_index, lambda sink_index: sink_index )
	source_mute = _pulse_method_call(
		c.pa.context_set_source_mute_by_index, lambda mute=True: mute )
	source_output_volume_set = _pulse_method_call(
		c.pa.context_set_source_output_volume, lambda vol: vol.to_struct() )
	source_volume_set = _pulse_method_call(
		c.pa.context_set_source_volume_by_index, lambda vol: vol.to_struct() )
	source_suspend = _pulse_method_call(
		c.pa.context_suspend_source_by_index, lambda suspend=True: suspend )
	source_port_set = _pulse_method_call(
		c.pa.context_set_source_port_by_index,
		lambda port: port.name if isinstance(port, PulsePortInfo) else port )


	async def module_load(self, name, args=''):
		if is_list(args): args = ' '.join(args)
		name, args = map(c.force_bytes, [name, args])
		data = list()
		async with _pulse_op_cb(self, raw=True) as cb:
			cb = c.PA_CONTEXT_INDEX_CB_T(
				lambda ctx, index, userdata, cb=cb: data.append(index) or cb() )
			try: c.pa.context_load_module(self._ctx, name, args, cb, None)
			except c.pa.CallError as err: raise PulseOperationInvalid(err.args[-1])
		index, = data
		if index == c.PA_INVALID:
			raise PulseError('Failed to load module: {} {}'.format(name, args))
		return index

	module_unload = _pulse_method_call(c.pa.context_unload_module, None)


	async def stream_restore_test(self):
		'Returns module-stream-restore version int (e.g. 1) or None if it is unavailable.'
		data = list()
		async with _pulse_op_cb(self, raw=True) as cb:
			cb = c.PA_EXT_STREAM_RESTORE_TEST_CB_T(
				lambda ctx, version, userdata, cb=cb: data.append(version) or cb() )
			try: c.pa.ext_stream_restore_test(self._ctx, cb, None)
			except c.pa.CallError as err: raise PulseOperationInvalid(err.args[-1])
		version, = data
		return version if version != c.PA_INVALID else None

	stream_restore_read = _pulse_get_list(
		c.PA_EXT_STREAM_RESTORE_READ_CB_T,
		c.pa.ext_stream_restore_read, PulseExtStreamRestoreInfo, index_arg=False )
	stream_restore_list = stream_restore_read # for consistency with other *_list methods

	@ft.partial(_pulse_method_call, c.pa.ext_stream_restore_write, index_arg=False)
	def stream_restore_write( obj_name_or_list,
			mode='merge', apply_immediately=False, **obj_kws ):
		'''Update module-stream-restore db entry for specified name.
			Can be passed PulseExtStreamRestoreInfo object or list of them as argument,
				or name string there and object init keywords (e.g. volume, mute, channel_list, etc).
			"mode" is PulseUpdateEnum value of
				'merge' (default), 'replace' or 'set' (replaces ALL entries!!!).'''
		mode = PulseUpdateEnum[mode]._c_val
		if is_str(obj_name_or_list):
			obj_name_or_list = PulseExtStreamRestoreInfo(obj_name_or_list, **obj_kws)
		if isinstance(obj_name_or_list, PulseExtStreamRestoreInfo):
			obj_name_or_list = [obj_name_or_list]
		# obj_array is an array of structs, laid out contiguously in memory, not pointers
		obj_array = (c.PA_EXT_STREAM_RESTORE_INFO * len(obj_name_or_list))()
		for n, obj in enumerate(obj_name_or_list):
			obj_struct, dst_struct = obj.to_struct(), obj_array[n]
			for k,t in obj_struct._fields_: setattr(dst_struct, k, getattr(obj_struct, k))
		return mode, obj_array, len(obj_array), int(bool(apply_immediately))

	@ft.partial(_pulse_method_call, c.pa.ext_stream_restore_delete, index_arg=False)
	def stream_restore_delete(obj_name_or_list):
		'''Can be passed string name,
			PulseExtStreamRestoreInfo object or a list of any of these.'''
		if is_str(obj_name_or_list, PulseExtStreamRestoreInfo):
			obj_name_or_list = [obj_name_or_list]
		name_list = list((obj.name if isinstance( obj,
			PulseExtStreamRestoreInfo ) else obj) for obj in obj_name_or_list)
		name_struct = (c.c_char_p * len(name_list))()
		name_struct[:] = list(map(c.force_bytes, name_list))
		return [name_struct]


	async def default_set(self, obj):
		'Set passed sink or source to be used as default one by pulseaudio server.'
		assert_pulse_object(obj)
		method = {
			PulseSinkInfo: self.sink_default_set,
			PulseSourceInfo: self.source_default_set }.get(type(obj))
		if not method: raise NotImplementedError(type(obj))
		await method(obj)

	async def sink_default_get(self):
		'Wrapper around server_info() to return sink for default_sink_name there.'
		return await self.get_sink_by_name((await self.server_info()).default_sink_name)
	async def source_default_get(self):
		'Wrapper around server_info() to return source for default_source_name there.'
		return await self.get_source_by_name((await self.server_info()).default_source_name)

	async def mute(self, obj, mute=True):
		assert_pulse_object(obj)
		method = {
			PulseSinkInfo: self.sink_mute,
			PulseSinkInputInfo: self.sink_input_mute,
			PulseSourceInfo: self.source_mute,
			PulseSourceOutputInfo: self.source_output_mute }.get(type(obj))
		if not method: raise NotImplementedError(type(obj))
		await method(obj.index, mute)
		obj.mute = mute

	async def port_set(self, obj, port):
		assert_pulse_object(obj)
		method = {
			PulseSinkInfo: self.sink_port_set,
			PulseSourceInfo: self.source_port_set }.get(type(obj))
		if not method: raise NotImplementedError(type(obj))
		await method(obj.index, port)
		obj.port_active = port

	async def card_profile_set(self, card, profile):
		assert_pulse_object(card)
		if is_str(profile):
			profile_dict = dict((p.name, p) for p in card.profile_list)
			if profile not in profile_dict:
				raise PulseIndexError( 'Card does not have'
					' profile with specified name: {!r}'.format(profile) )
			profile = profile_dict[profile]
		await self.card_profile_set_by_index(card.index, profile.name)
		card.profile_active = profile

	async def volume_set(self, obj, vol):
		assert_pulse_object(obj)
		method = {
			PulseSinkInfo: self.sink_volume_set,
			PulseSinkInputInfo: self.sink_input_volume_set,
			PulseSourceInfo: self.source_volume_set,
			PulseSourceOutputInfo: self.source_output_volume_set }.get(type(obj))
		if not method: raise NotImplementedError(type(obj))
		await method(obj.index, vol)
		obj.volume = vol

	async def volume_set_all_chans(self, obj, vol):
		assert_pulse_object(obj)
		obj.volume.value_flat = vol
		await self.volume_set(obj, obj.volume)

	async def volume_change_all_chans(self, obj, inc):
		assert_pulse_object(obj)
		obj.volume.values = [max(0, v + inc) for v in obj.volume.values]
		await self.volume_set(obj, obj.volume)

	async def volume_get_all_chans(self, obj):
		# Purpose of this func can be a bit confusing, being here next to set/change ones
		'''Get "flat" volume float value for info-object as a mean of all channel values.
			Note that this DOES NOT query any kind of updated values from libpulse,
				and simply returns value(s) stored in passed object, i.e. same ones for same object.'''
		assert_pulse_object(obj)
		return obj.volume.value_flat

	async def _event_mask_set(self, *masks):
		mask = 0
		for m in masks: mask |= PulseEventMaskEnum[m]._c_val
		async with _pulse_op_cb(self) as cb:
			c.pa.context_subscribe(self._ctx, mask, cb, None)

	async def subscribe_events(self, *masks) -> AsyncIterator[PulseEventInfo]:
		'''Subscribes to PulseAudio events with the given event masks and creates an asynchronous
				iterator, yielding the events as they are received from the server.
			This method is an alternative to `event_callback_set` and `event_mask_set`.

			Raises a `PulseDisconnect` exception when the connection to the server is lost.
			Raises StopIteration (returns silently from for loop) when the event subscription is
				cancelled via `event_callback_set(None)`'''
		if self.event_callback is not None:
			raise RuntimeError('Only a single subscribe_events generator can be used at a time.')
		queue = asyncio.Queue()
		self.event_callback = queue.put_nowait
		try:
			await self._event_mask_set(*masks)
			while True:
				yield await self._wait_disconnect_or(queue.get())
		finally:
			self.event_callback = None
			if self._connected.is_set():
				await self._event_mask_set('null')

	async def get_peak_sample(self, source, timeout, stream_idx=None):
		'''Returns peak (max) value in 0-1.0 range for samples in source/stream within timespan.
			"source" can be either int index of pulseaudio source
				(i.e. source.index), its name (source.name), or None to use default source.
			Resulting value is what pulseaudio returns as
				PA_SAMPLE_FLOAT32NE float after "timeout" seconds.
			If specified source does not exist, 0 should be returned after timeout.
			This can be used to detect if there's any sound
				on the microphone or any sound played through a sink via its monitor_source index,
				or same for any specific stream connected to these (if "stream_idx" is passed).
			Sample stream masquerades as
				application.id=org.PulseAudio.pavucontrol to avoid being listed in various mixer apps.
			Example - get peak for specific sink input "si" for 0.8 seconds:
				await pulse.get_peak_sample(await pulse.sink_info(si.sink).monitor_source, 0.8, si.index)'''
		samples = [0.0]

		async def subscriber():
			async for volume in self.subscribe_peak_sample(source, 25, stream_idx):
				samples[0] = max(samples[0], volume)

		task = asyncio.get_event_loop().create_task(subscriber())
		try:
			await asyncio.wait_for(task, timeout)
		except asyncio.TimeoutError:
			task.cancel()
			with suppress(asyncio.CancelledError):
				await task

		return min(1.0, samples[0])

	async def subscribe_peak_sample(self, source, rate=25, stream_idx=None, allow_suspend=False
			) -> AsyncIterator[float]:
		"""
		Subscribe to a (downsampled) audio stream to monitor audio volume.

		Using PulseAudio's `stream_connect_record` method, the stream can either be a normal record stream from a
		source, a stream from a sink monitor source or a monitor stream of a specific sink input. To monitor the volume,
		we use the PA_SAMPLE_FLOAT32NE sample format. This method returns an asynchronous generator, which yields the
		volume samples as they are streamed from the PulseAudio server.

		Example usage for monitoring a source (e.g. microphone input) with 5Hz::

		  async for volume in pulse.subscribe_peak_sample(source.name, rate=5):
		      print("volume =", volume)

		Example usage for monitoring a sink input (i.e. application output):

		  async for volume in pulse.subscribe_peak_sample((await pulse.sink_info(sink_input.sink)).monitor_source,
		                                                  stream_idx=sink_input.index):
		      print("volume = ", volume)

		:param source: Name (!) of the source to monitor its volume. Use `PulseSinkInfo.monitor_source` to get the
			correct source name for monitoring a sink or an input of that sink.
		:param rate: Sample rate, i.e. rate of volume measurements yielded by the generator in 1/second
		:param stream_idx: When `source` is a sink monitor source, specify the index (!) of a sink input, to monitor
			this single sink input stream instead of the sink sum signal.
		:param allow_suspend: If True, the flat DONT_INHIBIT_AUTO_SUSPEND is set on the stream, such that Pulse Audio
			will automatically suspend the source or sink after some seconds, despite our monitor stream running. This
			is useful for monitoring sinks, but prevents actively monitoring sources for more than a few seconds.
		"""
		proplist = c.pa.proplist_from_string('')
		ss = c.PA_SAMPLE_SPEC(format=c.PA_SAMPLE_FLOAT32NE, rate=rate, channels=1)
		s = c.pa.stream_new_with_proplist(self._ctx, 'peak detect', c.byref(ss), None, proplist)
		queue = asyncio.Queue()
		c.pa.proplist_free(proplist)

		@c.PA_STREAM_REQUEST_CB_T
		def read_cb(s, nbytes, _userdata):
			bs = c.c_int(nbytes)
			buff = c.c_void_p()
			c.pa.stream_peek(s, buff, c.byref(bs))
			try:
				if not buff or bs.value < 4:
					return
				queue.put_nowait(c.cast(buff, c.POINTER(c.c_float))[0])
			finally:
				# stream_drop() flushes buffered data (incl. buff=NULL "hole" data)
				# stream.h: "should not be called if the buffer is empty"
				if bs.value:
					c.pa.stream_drop(s)

		if stream_idx is not None:
			c.pa.stream_set_monitor_stream(s, stream_idx)
		c.pa.stream_set_read_callback(s, read_cb, None)
		if source is not None:
			source = unicode(source).encode('utf-8')

		flags = c.PA_STREAM_DONT_MOVE | c.PA_STREAM_PEAK_DETECT | c.PA_STREAM_ADJUST_LATENCY
		if allow_suspend:
			flags |= c.PA_STREAM_DONT_INHIBIT_AUTO_SUSPEND
		try:
			c.pa.stream_connect_record(
				s, source,
				c.PA_BUFFER_ATTR(fragsize=4, maxlength=2**32-1),
				flags)
		except c.pa.CallError:
			c.pa.stream_unref(s)
			raise

		try:
			while True:
				yield await self._wait_disconnect_or(queue.get())
		finally:
			try:
				c.pa.stream_disconnect(s)
			except c.pa.CallError:
				pass  # stream was removed
			c.pa.stream_unref(s)

	async def play_sample(self, name, sink=None, volume=1.0, proplist_str=None):
		'''Play specified sound sample,
				with an optional sink object/name/index, volume and proplist string parameters.
			Sample must be stored on the server in advance, see e.g. "pacmd list-samples".
			See also libcanberra for an easy XDG theme sample loading, storage and playback API.'''
		if isinstance(sink, PulseSinkInfo): sink = sink.index
		sink = str(sink) if sink is not None else None
		proplist = c.pa.proplist_from_string(proplist_str) if proplist_str else None
		volume = int(round(volume*c.PA_VOLUME_NORM))
		async with _pulse_op_cb(self) as cb:
			try:
				if not proplist:
					c.pa.context_play_sample(self._ctx, name, sink, volume, cb, None)
				else:
					c.pa.context_play_sample_with_proplist(
						self._ctx, name, sink, volume, proplist, cb, None )
			except c.pa.CallError as err: raise PulseOperationInvalid(err.args[-1])



import asyncio
import signal
from contextlib import suppress

import pulsectl_asyncio

# import pulsectl
# print('Event types:', pulsectl.PulseEventTypeEnum)
# print('Event facilities:', pulsectl.PulseEventFacilityEnum)
# print('Event masks:', pulsectl.PulseEventMaskEnum)


async def listen():
    async with pulsectl_asyncio.PulseAsync('event-printer') as pulse:
        async for event in pulse.subscribe_events('all'):
            print('Pulse event:', event)


async def main():
    # Run listen() coroutine in task to allow cancelling it
    listen_task = loop.create_task(listen())

    # Schedule listen_task to be cancelled after 10 seconds
    # Alternatively, the PulseAudio event subscription can be ended by breaking/returning from the `async for` loop
    loop.call_later(5, listen_task.cancel)

    # register signal handlers to cancel listener when program is asked to terminate
    for sig in (signal.SIGTERM, signal.SIGHUP, signal.SIGINT):
        loop.add_signal_handler(sig, listen_task.cancel)

    with suppress(asyncio.CancelledError):
        await listen_task


# Run event loop until main_task finishes
loop = asyncio.get_event_loop()
loop.run_until_complete(main())



import asyncio
import pulsectl_asyncio

async def main():
    async with pulsectl_asyncio.PulseAsync('volume-increaser') as pulse:
        for sink in await pulse.sink_list():
            await pulse.volume_change_all_chans(sink, 0.1)

loop = asyncio.get_event_loop()
loop.run_until_complete(main())



import asyncio
import signal
from contextlib import suppress

import pulsectl_asyncio


async def listen(pulse: pulsectl_asyncio.PulseAsync, source_name: str):
    async for level in pulse.subscribe_peak_sample(source_name, rate=5):
        print('\x1b[2K\x1b[0E', end='')  # return to beginning of line
        num_o = round(level * 80)
        print('O' * num_o + '-' * (80-num_o), end='', flush=True)


async def main():
    """
    Monitor output level of the default sink.
    """
    async with pulsectl_asyncio.PulseAsync('peak-listener') as pulse:
        # Get name of monitor_source of default sink
        server_info = await pulse.server_info()
        default_sink_info = await pulse.get_sink_by_name(server_info.default_sink_name)
        source_name = default_sink_info.monitor_source_name

        # Start listening/monitoring task
        listen_task = loop.create_task(listen(pulse, source_name))

        # Schedule listen_task to be cancelled after 10 seconds
        # Alternatively, the PulseAudio event subscription can be ended by breaking/returning from the `async for` loop
        loop.call_later(5, listen_task.cancel)

        # register signal handlers to cancel listener when program is asked to terminate
        for sig in (signal.SIGTERM, signal.SIGHUP, signal.SIGINT):
            loop.add_signal_handler(sig, listen_task.cancel)

        with suppress(asyncio.CancelledError):
            await listen_task
            print()


# Run event loop until main_task finishes
loop = asyncio.get_event_loop()
loop.run_until_complete(main())
